/*
 * NAME:
 *	net_read,net_write - interface stdio to read_net and write_net
 *
 * SYNOPSIS:
 *	#include <netio.h>
 *
 *	int	net_read(int fd, char *buf, int nbytes);
 *	int	net_write(int fd, char *buf, int nbytes);
 *
 *
 * DESCRIPTION:
 *	These routines simply allow encryption hooks to be used with
 *	applications that use stdio.
 *
 *	'netio.h' contains macro definitions which override fprintf(),
 *	vfprintf(), fgets(), fgetc(), getc(), putc() and fputc().
 *
 *	Each of the net_*() functions, checks whether the FILE * is
 *	associated with an SSL connection and if not calls the normal
 *	stdio function, otherwise it implements the appropriate
 *	buffered I/O behaviour with net_filbuf() and net_flsbuf()
 *	calling read_net() and write_net() respectively.
 *	
 * SEE ALSO:
 *	rwnet.c
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 *	
 *	This is a re-work of a stdio to socket implementation for DOS
 *	I did back in the late 1980s.
 */
/*
 *      @(#)Copyright (c) 1989,1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: netrw.c,v 1.9 1997/05/16 08:17:19 sjg Exp $";
#endif

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#ifdef USE_SSL
# include "ssl.h"
# include "sslfd.h"
#endif

#include "netiob.h"
#ifdef NEED_NETIOB
# define _IOB_RCNT _cnt
# define _IOB_WCNT _cnt
# define _IOB_PTR _ptr
# define _IOB_BASE _base
# define _IOB_FLAG _flag
# define _IOB_BUFSIZ _bufsiz
# define _IOB_FILE _file
#else
# include "stio.h"
#endif

/*
 * Ok, this is perhaps getting silly,
 * but these provide read(2) and write(2) semantics without breaking netio.
 * Be sure to call net_getiob(fd, 1, w) to ensure a net_iob struct is
 * setup before * calling any of getc(), fgetc(), fgets() etc. 
 */

/*
 * If we do not have a net_iob set up, we call _read(),
 * or if it is empty, we call net_getc(),net_ungetc() to populate it, 
 * then we grab up to nbytes from the net_iob.
 * All the mucking about is to try and ensure we do a max of 1 I/O.
 */

int
net_read(fd, buf, nbytes)
	int fd;
	unsigned char *buf;
	int nbytes;
{
	NETIOB *p = net_getiob(fd, 0, 0);
	int n;
	
	if (!p)
		return _read(fd, buf, nbytes);

	if (p->_IOB_RCNT <= 0 || p->_IOB_PTR == p->_IOB_BASE) {	/* empty */
		/*
		 * then this should do no harm?
		 */
		if (nbytes > 512)
			return _read(fd, buf, nbytes);
		/*
		 * The bugger may be calling read(fd, buf, 1)!
		 */
		if ((n = net_getc(p)) == EOF)
			return EOF;
		_net_ungetc(n, p);
	}
	n = MIN(nbytes, p->_IOB_RCNT);
#ifdef _IOB_BUFSIZ
	n = MIN(n, p->_IOB_BUFSIZ);
#endif
	memcpy(buf, p->_IOB_PTR, n);
	p->_IOB_RCNT -= n;
	if (p->_IOB_RCNT <= 0) {
		p->_IOB_RCNT = 0;
		p->_IOB_PTR = p->_IOB_BASE;
	} else {
		p->_IOB_PTR += n;
	}
	return n;
}


/*
 * if we have a net_iob, flush it, then return _write()
 */
int
net_write(fd, buf, nbytes)
	int fd;
	unsigned char *buf;
	int nbytes;
{
	NETIOB *p = net_getiob(fd, 0, 1);
	int n;

	
#if 0
	/*
	 * this saves a copy
	 */
	if (p) {
		if (_net_fflush(p) == EOF)
			return EOF;
	}
	return _write(fd, buf, nbytes);
#else
	/*
	 * this might be more network friendly
	 */
	if (!p || p->_IOB_PTR == p->_IOB_BASE)	/* nothing to flush */
		return _write(fd, buf, nbytes);
	for (n = 0; n < nbytes; n++)
		if (net_putc(buf[n], p) == EOF)
			return EOF;
	if (_net_fflush(p) == EOF)
		return EOF;
	return n;
#endif
}

#ifdef MAIN
int
main(argc, argv)
	int argc;
	char **argv;
{
	char buf[1024];
	int c, n, x, i;
	int done = 0;
	
	if (argc > 1) {
		if ((c = net_getc(stdin)) == EOF)
			exit(1);
		net_putc(c, stdout);
	}
	while (!done) {
		if (argc > 2) {
			if ((c = net_getc(stdin)) == EOF) {
				done = 1;
				break;
			}
			net_putc(c, stdout);
		}
		for (i = 1, x = n =  0; i < sizeof(buf); n += c, i *= 4)
			if ((c = net_read(0, &buf[n], i)) <= 0) {
				done = 1;
				break;
			}
		for (x = 0; x < n; x += c) {
			c = net_write(1, &buf[x], n - x);
		}
	}
	exit(0);
}
#endif
