/*
 * Copyright (c) 1996 Simon J. Gerraty, all rights reserved.
 * Copyright (c) 1983, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char RCSid[] = "$Id: sslrcmd.c,v 1.22 1998/08/12 13:11:21 sjg Exp $";
#endif

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <signal.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifndef SA_NOCLDSTOP
# include "sigact.h"				/* <sjg> fake POSIX signals */
#endif

#ifdef sun
# include "sunos.h"
#endif

#include "sslfd.h"

#ifdef SOCKS
# include "socks.h"
#endif

#ifndef SSHELL_PORT
# define SSHELL_PORT	614			/* official port */
#endif


/*
 * <sjg> simplified rcmd()
 * We do not use suid, reserved ports etc,
 * since our cert proves who we are.
 */
int
ssl_rcmd(ahost, rport, locuser, remuser, cmd, fd2p)
	char **ahost;
	u_short rport;
	const char *locuser, *remuser, *cmd;
	int *fd2p;
{
	extern char *use_proxy();
	char dest[256];
	const char *host, *proxy = 0;
	struct hostent *hp;
	struct sockaddr_in sin, from;
	fd_set reads;
	sigset_t mask, omask;
	pid_t pid;
	int s, lport = 0, timo;
	char c;
	
	pid = getpid();

	if (rport == 0 || rport == (u_short)-1) {
		struct servent *sp;

		if ((sp = getservbyname("sshell", "tcp")) || /* official */
		    (sp = getservbyname("SSLshell", "tcp"))) { /* old name */
			rport = sp->s_port;
		} else {
			rport = htons(SSHELL_PORT);
		}
	}

	sigemptyset(&mask);
	sigaddset(&mask,SIGURG);
	(void) sigprocmask(SIG_BLOCK, &mask, &omask);

	/*
	 * decide on proxy, _before_ we qualify *ahost
	 */
	if (ssl_proxy)
		proxy = ssl_proxy;
	else
		proxy = use_proxy(*ahost, "sshell");

	if (proxy && *proxy) {
		if (fd2p)
			*fd2p = -1;		/* make sure its not valid */
		fd2p = 0;			/* not supported */
		host = proxy;
	} else {
		proxy = 0;		/* incase it was empty string */
		host = dest;
	}

	hp = gethostbyname(*ahost);
	if (hp)
		*ahost = hp->h_name;	/* why? */
	else if (proxy == 0) {
#if defined(BSD4_4) || defined(HAVE_HERROR)
		herror(*ahost);
#else
		fprintf(stderr, "ssl_rcmd: host '%s' unknown, or lookup failed.\n", *ahost);
#endif
		return (-1);
	}
	
	sprintf(dest, "%.200s:%d", *ahost, ntohs(rport));

	if ((s = net_open(host, "c")) < 0) {
		if (errno == 0) {
#if defined(BSD4_4) || defined(HAVE_HERROR)
			fprintf(stderr, "ssl_rcmd: ");
			herror(host);
#else
			fprintf(stderr, "ssl_rcmd: host '%s' unknown, or lookup failed.\n", host);
#endif
		} else {
			(void)fprintf(stderr, "%s: %s\n", *ahost, strerror(errno));
		}
		(void) sigprocmask(SIG_SETMASK, &omask, 0);
		return (-1);
	}

#ifdef F_SETOWN
	fcntl(s, F_SETOWN, pid);
#endif
	if (proxy) {
		char *err;
		
		if ((s = proxy_connect(s, dest, &err)) < 0) {
			(void)fprintf(stderr, "ssl_rcmd: proxy_connect failed %d\n%s", -s, err);
			return (-1);
		}
		/*
		 * now connected
		 */
	}

	if (ssl_start(s, -1, 0) < 0)
		goto bad;
	
	if (fd2p == 0) {
		write_net(s, "", 1);
		lport = 0;
	} else {
		/* <sjg> all quite different */
		char num[8];
		int nfd, s3, s2 = socket(AF_INET, SOCK_STREAM, 0);
		int len = sizeof(from);

		if (s2 < 0)
			goto bad;
		listen(s2, 1);
		bzero((char *)&from, sizeof(from));
		from.sin_family = AF_INET;
		from.sin_port = 0;
		if (getsockname(s2, (struct sockaddr *)&from, &len) < 0) {
			perror("getsockname(s2)");
			close(s2);
			goto bad;
		}
		lport =  ntohs(from.sin_port);
		(void)sprintf(num, "%d", lport);
		if (write_net(s, num, strlen(num)+1) != strlen(num)+1) {
			(void)fprintf(stderr,
			    "ssl_rcmd: write (setting up stderr): %s\n",
			    strerror(errno));
			(void)close(s2);
			goto bad;
		}
		FD_ZERO(&reads);
		FD_SET(s, &reads);
		FD_SET(s2, &reads);
		errno = 0;
		nfd = (s > s2) ? s : s2;
		if (select(nfd + 1, &reads, 0, 0, 0) < 1 ||
		    !FD_ISSET(s2, &reads)) {
			if (errno != 0)
				(void)fprintf(stderr,
				    "ssl_rcmd: select (setting up stderr): %s\n",
				    strerror(errno));
			else
				(void)fprintf(stderr,
				"select: protocol failure in circuit setup\n");
			(void)close(s2);
			goto bad;
		}
		len = sizeof(from);
		s3 = accept(s2, (struct sockaddr *)&from, &len);
		(void)close(s2);
		if (s3 < 0) {
			(void)fprintf(stderr,
			    "ssl_rcmd: accept: %s\n", strerror(errno));
			lport = 0;
			goto bad;
		}
		*fd2p = s3;
		if (ssl_clone(s3, s, -1, 0) < 0) /* uses previous setup */
			goto bad2;
	}
	
	(void)write_net(s, (char *) locuser, strlen(locuser)+1);
	(void)write_net(s, (char *) remuser, strlen(remuser)+1);
	(void)write_net(s, (char *) cmd, strlen(cmd)+1);

	if (read_net(s, &c, 1) != 1) {
		(void)fprintf(stderr,
		    "ssl_rcmd: %s: %s\n", *ahost, strerror(errno));
		goto bad2;
	}
	if (c != 0) {
		char buf[BUFSIZ];
		int i;
		
		if ((i = read_net(s, buf, sizeof(buf))) > 0) {
			(void)write(STDERR_FILENO, buf, i);
		}
		goto bad2;
	}
	(void) sigprocmask(SIG_SETMASK, &omask, 0);
	return (s);
bad2:
	if (lport)
		(void)ssl_close(*fd2p);
bad:
	(void)close(s);
	(void) sigprocmask(SIG_SETMASK, &omask, 0);
	return (-1);
}
